
import cherrypy
import base64
import hashlib
import json
import os

import api


class Instance (api.API):
    def __init__(self):
        self._users = json.load(open("users.json"))

    def _check_credentials(self, username, password):
        password = hashlib.sha512(password.encode()).hexdigest()
        return username in self._users and self._users[username] == password

    def _validate(self, auth):
        if auth.startswith("Basic "):
            auth = auth[6:].encode()
            username, password = base64.b64decode(auth).decode().split(":")
            return self._check_credentials(username, password)
        return False

if __name__ == '__main__':
    cwd = os.path.split(os.path.abspath(__file__))[0]
    # you can make this next line into a comment when testing things
    #cherrypy.config.update({'engine.autoreload_on': False})
    cherrypy.quickstart(Instance(), config={'global': {
        'tools.encode.encoding': 'utf8',
        'server.socket_host': '0.0.0.0',
        'server.socket_port': 9001,
        'server.ssl_certificate': '/etc/ssl/certs/ssl-cert-snakeoil.pem',
        'server.ssl_private_key': '/etc/ssl/private/ssl-cert-snakeoil.key',
        'log.access_file': '/var/log/sentry/access.log',
        'log.error_file': '/var/log/sentry/error.log',
        'tools.gzip.on': True,
    }})




import copy
import yaml

TPLS = {
    'juju-core': {
        'machines': {
            'agent-state': 'started',
            'agent-version': '1.11.0',
            'dns-name': 'xxx',
            'instance-id': 'i-xxx',
            'series': 'precise',
        }, 'services': {
            'charm': 'cs:precise/%s-x',
            'exposed': False,
        }, 'units': {
            'agent-state': 'pending',
            'agent-version': '1.11.0',
            'machine': '',
            'public-address': 'xxx',
        }
    }, 'juju': {
        'machines': {
            'agent-state': 'running',
            'dns-name': 'xxx',
            'instance-id': 'i-xxx',
            'instance-state': 'running',
        }, 'services': {
            'charm': 'cs:precise/%s-x',
            'relations': {},
        }, 'units': {
            'agent-state': 'pending',
            'public-address': 'xxx',
            'machine': '',
        }
    }
}


class JujuStatus(object):
    def __init__(self, version='juju-core'):
        self.status = {'machines': {}, 'services': {}}
        self.version = version
        self.add_machine()

    def add(self, service, state='started', relation_error=None, dying=False):
        if not service in self.status['services']:
            stpl = self.get_template('services')
            stpl['charm'] = stpl['charm'] % service
            stpl['units'] = {}
            self.status['services'][service] = stpl

        if dying and self.version == 'juju-core':
            self.status['services'][service]['life'] = 'dying'

        machine = self.add_machine()
        unit = "%s/%s" % (service,
                          len(self.status['services'][service]['units']))
        utpl = self.get_template('units')
        utpl['machine'] = machine
        utpl['agent-state'] = state

        if relation_error:
            if self.version == 'juju':
                utpl['relation-error'] = relation_error
            else:
                utpl['state'] = 'error %s' % relation_error

        self.status['services'][service]['units'][unit] = utpl

    def add_machine(self):
        tpl = self.get_template('machines')
        machine = len(self.status['machines'])
        if self.version == 'juju-core':
            machine = str(machine)

        self.status['machines'][machine] = tpl
        return machine

    def get_template(self, key, version=None):
        global TPLS
        if not version:
            version = self.version

        return copy.deepcopy(TPLS[version][key])

    def __str__(self):
        return yaml.dump(self.status, default_flow_style=False)



from __future__ import print_function

import os
import sys
import yaml
import signal
import subprocess
import errno

from contextlib import contextmanager

SKIP = 100
PASS = 0
FAIL = 1


class TimeoutError(Exception):
    def __init__(self, value="Timed Out"):
        self.value = value


def _as_text(bytestring):
    """Naive conversion of subprocess output to Python string"""
    return bytestring.decode("utf-8", "replace")


def setup_bzr(charm_dir):
    try:
        run_bzr(['whoami'], charm_dir)
    except IOError:
        run_bzr(['whoami', 'amulet@dummy-user.tld'], charm_dir)

    run_bzr(["init"], charm_dir)


def run_bzr(args, working_dir, env=None):
    """Run a Bazaar command in a subprocess"""
    try:
        p = subprocess.Popen(["bzr"] + args, cwd=working_dir, env=env,
                             stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    except OSError as e:
        if e.errno != errno.ENOENT:
            raise
        raise OSError("bzr not found, do you have Bazaar installed?")
    out, err = p.communicate()
    if p.returncode:
        raise IOError("bzr command failed {!r}:\n"
                      "{}".format(args, _as_text(err)))
    return _as_text(out)


def juju(args, env=None):
    try:
        p = subprocess.Popen(['juju'] + args, env=env, stdout=subprocess.PIPE,
                             stderr=subprocess.PIPE)
    except OSError as e:
        if e.errno != errno.ENOENT:
            raise
        raise OSError("juju not found, do you have Juju installed?")
    out, err = p.communicate()
    if p.returncode:
        raise IOError("juju command failed {!r}:\n"
                      "{}".format(args, _as_text(err)))
    return _as_text(out) if out else None


@contextmanager
def timeout(seconds):
    def signal_handler(signum, frame):
        raise TimeoutError()
    signal.signal(signal.SIGALRM, signal_handler)
    signal.alarm(seconds)
    try:
        yield
    finally:
        signal.alarm(0)


class JujuVersion(object):
    def __init__(self, major=0, minor=0, patch=0, get_version=True):
        self.mapping = ['major', 'minor', 'patch']
        self.major = major
        self.minor = minor
        self.patch = patch

        if get_version:
            self.get_version()

    def parse_version(self, version_str):
        version = version_str.split()
        if len(version) > 1:
            version_str = str(version[1])
        else:
            version_str = str(version[0])

        return version_str.split('-')[0].split('.')

    def update_version(self, version_list):
        for i, ver in enumerate(version_list):
            try:
                setattr(self, self.mapping[i], int(ver))
            except:
                break  # List out of range? Versions not semantic? Too bad

    def get_version(self):
        try:
            version = juju(['version'])
        except:
            version = juju(['--version'])

        self.update_version(self.parse_version(version))

    def __str__(self):
        return '.'.join(str(v) for v in [self.major, self.minor, self.patch])


def environments(juju_home="~/.juju/"):
    env_file = os.path.expanduser(os.path.join(juju_home, 'environments.yaml'))
    if not os.path.isfile(env_file):
        raise IOError('%s was not found.' % env_file)

    with open(env_file, 'r') as env_yaml:
        envs = yaml.safe_load(env_yaml.read())

    return envs


def raise_status(code, msg=None):
    if msg:
        print(msg)

    sys.exit(code)


def default_environment(juju_home="~/.juju/"):
    juju_home = os.path.expanduser(juju_home)
    envs = environments(juju_home)

    if 'JUJU_ENV' in os.environ:
        return os.environ['JUJU_ENV']

    if os.path.exists(os.path.join(juju_home, 'current-environment')):
        cur_env = None
        with open(os.path.join(juju_home, 'current-environment')) as f:
            cur_env = f.read().strip()

        if cur_env in envs['environments']:
            return cur_env

    if 'default' in envs:
        return envs['default']
    else:
        if len(envs['environments']) != 1:
            raise ValueError('No default environment specified.')

        return next(iter(envs['environments'].keys()))




import os
import yaml
import glob
import shutil
import tempfile

from charmworldlib.charm import Charm
from .helpers import run_bzr, setup_bzr


def get_relation(charm, relation, cache=None):
    if cache and charm in cache:
        c = cache[charm]
    else:
        c = get_charm(charm)

    relations = c.relations

    if not relations:
        raise Exception('No relations for charm')

    for rel_type in relations:
        for rel_name in relations[rel_type]:
            if rel_name == relation:
                return rel_type, relations[rel_type][rel_name]['interface']

    return (None, None)


class Builder(object):
    def __init__(self, name, template, subordinate=False, hook='hooks.py'):
        self.metadata = {'name': name,
                         'summary': 'Generated by Amulet',
                         'description': 'Generated by Amulet',
                         'subordinate': subordinate,
                         'maintainer': 'Amulet Built <juju@lists.ubuntu.com>'}

        self.hook = hook

        self.template = os.path.realpath(template)
        if not os.path.exists(self.template):
            raise IOError('%s does not exist' % self.template)

        d = tempfile.mkdtemp(prefix='sentry%s_'
                             % ('-sub' if subordinate else ''))
        self.charm = os.path.join(d, name)
        shutil.copytree(self.template, self.charm, symlinks=True)
        for h in glob.glob(os.path.join(self.charm, 'hooks', '*')):
            os.chmod(h, 0o755)

        setup_bzr(self.charm)

        if subordinate:
            self.require('juju-info', 'juju-info', {'scope': 'container'})

        os.chmod(os.path.join(self.charm, 'hooks', self.hook), 0o755)

    def require(self, relation, interface, opts=None):
        if not opts:
            opts = {}
        opts['interface'] = interface
        self._add_relation('requires', relation, opts)

    def provide(self, relation, interface, opts=None):
        if not opts:
            opts = {}
        opts['interface'] = interface
        self._add_relation('provides', relation, opts)

    def peer(self, relation, interface, opts=None):
        if not opts:
            opts = {}
        opts['interface'] = interface
        self._add_relation('peers', relation, opts)

    def _add_relation(self, rel_type, relation, opts=None):
        if not rel_type in self.metadata:
            self.metadata[rel_type] = {}

        self.metadata[rel_type][relation] = opts

        # Build symlink to "global hooks" file
        if self.hook:
            for event in ['joined', 'changed', 'departed', 'broken']:
                hook_file = os.path.join(self.charm, 'hooks', '%s-relation-%s'
                                         % (relation, event))
                if not os.path.exists(hook_file):
                    os.symlink(self.hook, hook_file)

        self.write_metadata()

    def write_metadata(self):
        metadata = yaml.safe_dump(self.metadata, default_flow_style=False)
        with open(os.path.join(self.charm, 'metadata.yaml'), 'w') as m:
            m.write(metadata)

        self.save("Writing Metadata")

    def save(self, msg="Checkpoint"):
        run_bzr(["add", "."], self.charm)
        run_bzr(["commit", "--unchanged", "-m", msg], self.charm)


def get_charm(charm_path):
    if charm_path.startswith('cs:'):
        return Charm(charm_path)
    if charm_path.startswith('lp:'):
        return LaunchpadCharm(charm_path)
    if os.path.exists(os.path.expanduser(charm_path)):
        return LocalCharm(charm_path)

    return Charm(charm_path)


class LocalCharm(object):
    def __init__(self, path):
        path = os.path.abspath(os.path.expanduser(path))

        if not os.path.exists(os.path.join(path, 'metadata.yaml')):
            raise Exception('Charm not found')

        if not os.path.exists(os.path.join(path, '.bzr')):
            path = self._make_temp_copy(path)

        self.url = None
        self.subordinate = False
        self.relations = {}
        self.provides = {}
        self.requires = {}
        self.code_source = self.source = {'location': path}
        self._raw = self._load(os.path.join(path, 'metadata.yaml'))
        self._parse(self._raw)

    def _make_temp_copy(self, path):
        d = tempfile.mkdtemp(prefix='charm')
        temp_charm_dir = os.path.join(d, os.path.basename(path))
        shutil.copytree(path, temp_charm_dir, symlinks=True)
        setup_bzr(temp_charm_dir)
        run_bzr(["add", "."], temp_charm_dir)
        run_bzr(["commit", "--unchanged", "-m", "Copied from {}".format(path)],
                temp_charm_dir)
        self.temp_dir = d
        return temp_charm_dir

    def _parse(self, metadata):
        rel_keys = ['provides', 'requires']
        for key, val in metadata.items():
            if key in rel_keys:
                self.relations[key] = val

            setattr(self, key, val)

    def _load(self, metadata_path):
        with open(metadata_path) as f:
            data = yaml.safe_load(f.read())

        return data

    def __str__(self):
        return yaml.dump(self._raw)

    def __repr__(self):
        return '<LocalCharm %s>' % self.code_source['location']

    def __del__(self):
        temp_dir = getattr(self, 'temp_dir', None)
        if temp_dir:
            shutil.rmtree(temp_dir)


class LaunchpadCharm(object):
    def __init__(self, branch):
        self.url = None
        self.subordinate = False
        self.code_source = self.source = {'location': branch, 'type': 'bzr'}
        self.relations = {}
        self.provides = {}
        self.requires = {}
        self._raw = self._load(os.path.join(branch, 'metadata.yaml'))
        self._parse(self._raw)

    def _parse(self, metadata):
        rel_keys = ['provides', 'requires']
        for key, val in metadata.items():
            if key in rel_keys:
                self.relations[key] = val

            setattr(self, key, val)

    def _load(self, metadata_path):
        mdata = run_bzr(['cat', metadata_path], None)
        return yaml.safe_load(mdata)

    def __str__(self):
        return yaml.dump(self._raw)

    def __repr__(self):
        return '<LaunchpadCharm %s>' % self.code_source['location']




import os
import json
import copy
import base64
import shutil
import subprocess
import tempfile

from .helpers import default_environment, juju, timeout as unit_timesout
from .sentry import Talisman

from .charm import Builder, get_relation, get_charm

_default_sentry_template = os.path.join(
    os.path.abspath(os.path.dirname(__file__)), 'charms/sentry')


class CharmCache(dict):
    def __init__(self, test_charm):
        super(CharmCache, self).__init__()
        self.test_charm = test_charm

    def __getitem__(self, service):
        return self.fetch(service)

    def fetch(self, service, charm=None):
        try:
            return super(CharmCache, self).__getitem__(service)
        except KeyError:
            charm = charm or service
            self[service] = get_charm(
                os.getcwd() if charm == self.test_charm else charm)
            return super(CharmCache, self).__getitem__(service)


class Deployment(object):
    def __init__(self, juju_env=None, series='precise', sentries=True,
                 juju_deployer='juju-deployer',
                 sentry_template=None):
        self.services = {}
        self.relations = []
        self.interfaces = []
        self.subordinates = []
        self.series = series
        self.deployed = False
        self.juju_env = juju_env or default_environment()
        self.charm_name = os.path.basename(os.getcwd())

        self.sentry = None
        self._sentries = {}
        self.use_sentries = sentries
        self.sentry_blacklist = []
        self.sentry_template = sentry_template or _default_sentry_template
        self.relationship_sentry = None

        self.deployer = juju_deployer
        self.deployer_dir = tempfile.mkdtemp(prefix='amulet_deployment_')

        if 'JUJU_TEST_CHARM' in os.environ:
            self.charm_name = os.environ['JUJU_TEST_CHARM']

        self.charm_cache = CharmCache(self.charm_name)

    def load(self, deploy_cfg):
        schema = next(iter(deploy_cfg.values()))
        for service, service_config in schema['services'].items():
            self.add(
                service,
                charm=service_config.get('charm'),
                units=service_config.get('num_units', 1))
        self.series = schema['series']
        self.relations = schema['relations']

    def add(self, service, charm=None, units=1):
        if self.deployed:
            raise NotImplementedError('Environment already setup')
        subordinate = False
        if service in self.services:
            raise ValueError('Service is already set to be deployed')

        c = self.charm_cache.fetch(service, charm)

        if c.subordinate:
            subordinate = True
            for rtype in ['provides', 'requires']:
                try:
                    rels = getattr(c, rtype)
                    for relation in rels:
                        rel = rels[relation]
                        if 'scope' in rel and rel['scope'] == 'container':
                            self.subordinates.append('%s:%s' %
                                                     (service, relation))
                except:
                    pass

        if c.url:
            self.services[service] = {'charm': c.url}
        else:
            self.services[service] = {'branch': c.code_source['location']}
        if subordinate:
            self.services[service]['_has_sentry'] = True
        if units > 1:
            self.services[service]['num_units'] = units

    def add_unit(self, service, units=1):
        if not isinstance(units, int) or units < 1:
            raise ValueError('Only positive integers can be used for units')
        if service not in self.services:
            raise ValueError('Service needs to be added before you can scale')

        self.services[service]['num_units'] = \
            self.services[service].get('num_units', 1) + units

        if self.deployed:
            output = juju(['add-unit', service])
            if self.use_sentries:
                self.sentry = Talisman(self.services)
            return output

    def remove_unit(self, *args):
        if not self.deployed:
            raise NotImplementedError('Environment not setup yet')
        if not args:
            raise ValueError('No units provided')
        for unit in args:
            if '/' not in unit:
                raise ValueError('%s is not a unit' % unit)
            service = unit.split('/')[0]
            if service not in self.services:
                raise ValueError('%s, is not a deployed service' % unit)

        units = [unit for unit in args]
        juju(['remove-unit'] + units)

    def relate(self, *args):
        if len(args) < 2:
            raise LookupError('Need at least two services:relation')
        if self.deployed:
            raise NotImplementedError('Environment already setup')

        for srv_rel in args:
            if not ':' in srv_rel:
                raise ValueError('All relations must be explicit, ' +
                                 'service:relation')

            srv, rel = srv_rel.split(':')
            if srv not in self.services:
                raise ValueError('Can not relate, service not deployed yet')

            c = self.charm_cache[srv]
            if rel not in list(c.provides.keys()) + list(c.requires.keys()) \
                    + ['juju-info']:
                raise ValueError('%s does not exist for %s' % (rel, srv))

        args = list(args)
        first = args.pop(0)
        for srv in args:
            self._relate(first, srv)

    def _relate(self, a, b):
        self.relations.append([a, b])

    def unrelate(self, *args):
        if len(args) != 2:
            raise LookupError('Need exactly two service:relations')
        if not self.deployed:
            raise NotImplementedError('Environment not setup yet')

        for srv_rel in args:
            if not ':' in srv_rel:
                raise ValueError('All relations must be explicit, ' +
                                 'service:relation')

        sentry_relations = self._get_sentry_relations(*args)
        for relation in sentry_relations:
            juju(['remove-relation'] + relation)

    def _get_sentry_relations(self, service_rel_1, service_rel_2):
        """Return the two relations that join ``service_rel_1`` and
        ``service_rel_2`` via a relation sentry.

        """
        def _get_sentry_relation(rel_1, rel_2):
            for r in self.relations:
                if (r[1] == rel_2 and
                        r[0].startswith('relation-sentry') and
                        '_'.join(rel_1.split(':')) in r[0]):
                    return r
            raise LookupError(
                'Could not find relation between {} and {}'.format(
                    service_rel_1, service_rel_2))

        return [
            _get_sentry_relation(service_rel_1, service_rel_2),
            _get_sentry_relation(service_rel_2, service_rel_1),
        ]

    def schema(self):
        return self.deployer_map(self.services, self.relations)

    def configure(self, service, options):
        for k, v in options.items():
            include_token = 'include-base64://'
            if type(v) is str and v.startswith(include_token):
                v = v.replace(include_token, '')
                with open(os.path.join(os.getcwd(), 'tests', v)) as f:
                    v = base64.b64encode(f.read())
                service['options'][k] = v

        if self.deployed:
            opts = ['set', service]
            for k, v in options.items():
                opts.append("%s=%s" % (k, v))
            return juju(opts)

        if service not in self.services:
            raise ValueError('Service has not yet been described')
        if not 'options' in self.services[service]:
            self.services[service]['options'] = options
        else:
            self.services[service]['options'].update(options)

    def expose(self, service):
        if self.deployed:
            return juju(['expose', service])

        if service not in self.services:
            raise ValueError('%s has not yet been described' % service)
        self.services[service]['expose'] = True

    def setup(self, timeout=600, cleanup=True):
        """Deploy the workload.

        :param timeout: Amount of time to wait for deployment to complete.
        :param cleanup: Set to False to leave the generated deployer file
            on disk. Useful for debugging.

        """
        if not self.deployer:
            raise NameError('Path to juju-deployer is not defined.')

        _, s = tempfile.mkstemp(prefix='amulet-juju-deployer-', suffix='.json')
        with open(s, 'w') as f:
            f.write(json.dumps(self.schema()))

        try:
            with unit_timesout(timeout):
                subprocess.check_call([os.path.expanduser(self.deployer), '-W',
                                       '-L', '-c', s, '-e', self.juju_env,
                                       self.juju_env], cwd=self.deployer_dir)
            self.deployed = True
        except subprocess.CalledProcessError:
            raise
        finally:
            if cleanup:
                os.remove(s)

        if not self.deployed:
            raise Exception('Deployment failed for an unknown reason')

        if self.deployed and self.use_sentries:
            self.sentry = Talisman(self.services)

    def deployer_map(self, services, relations):
        if self.use_sentries:
            self.build_sentries()

        deployer_map = {
            self.juju_env: {
                'series': self.series,
                'services': self.services,
                'relations': self.build_relations()
            }
        }

        return deployer_map

    def build_relations(self):
        relations = []
        for rel in self.relations:
            relations.append(rel)

        return relations

    def build_sentries(self):
        services = copy.deepcopy(self.services)
        for service, details in services.items():
            if service in self._sentries:
                continue

            if not '_has_sentry' in details or not details['_has_sentry']:
                sentry = Builder('%s-sentry' % service, self.sentry_template,
                                 subordinate=True)
                self.add(sentry.metadata['name'], sentry.charm)
                self._relate('%s:juju-info' % service,
                             '%s:juju-info' % sentry.metadata['name'])
                self.expose(sentry.metadata['name'])
                self._sentries[sentry.metadata['name']] = sentry
                self.services[service]['_has_sentry'] = True
                charm = self.charm_cache[service]
                if hasattr(charm, 'series'):
                    self.services[sentry.metadata['name']]['series'] = charm.series

        # Build relationship sentry
        if not self.relationship_sentry:
            # Auto generate name
            rel_sentry = Builder('relation-sentry', self.sentry_template)
            rel_sentry.write_metadata()

            self.add(rel_sentry.metadata['name'], rel_sentry.charm)
            self.expose(rel_sentry.metadata['name'])
            self._sentries[rel_sentry.metadata['name']] = rel_sentry
            rel_sentry.write_metadata()
            self.relationship_sentry = rel_sentry

        relations = copy.deepcopy(self.relations)
        relation_sentry = self.relationship_sentry.metadata['name']
        for relation in relations:
            for rel in relation:
                if rel in self.subordinates:
                    break
                service, rel_name = rel.split(':')
                if service in self._sentries:
                    break
            else:
                relation_name = "-".join(relation).replace(':', '_')
                self.relations.remove(relation)
                try:
                    interface = get_relation(service, rel_name,
                                             self.charm_cache)[1]
                except:
                    continue

                if not interface:
                    raise Exception('Unable to detect interface for %s on %s'
                                    % (service, rel_name))

                self.relationship_sentry.provide('%s-%s' %
                                                 ('requires', relation_name),
                                                 interface)
                self.relationship_sentry.require('%s-%s' %
                                                 ('provides', relation_name),
                                                 interface)

                for rel in relation:
                    rel_data = get_relation(*rel.split(':'),
                                            cache=self.charm_cache)
                    self._relate('%s:%s-%s'
                                 % (relation_sentry, rel_data[0],
                                    relation_name), rel)

    def cleanup(self):
        shutil.rmtree(self.deployer_dir)
        for sentry in self._sentries:
            shutil.rmtree(os.path.dirname(self._sentries[sentry].charm))



import unittest
import yaml

from amulet.sentry import Talisman
from mock import patch


mock_status = yaml.load("""\
services:
  meteor:
    units:
      meteor/0:
        public-address: 10.0.3.152
      meteor/1:
        public-address: 10.0.3.177
  relation-sentry:
    units:
      relation-sentry/0:
        public-address: 10.0.3.92
""")


class TalismanTest(unittest.TestCase):

    @patch.object(Talisman, 'wait_for_status')
    @patch('amulet.sentry.helpers.default_environment')
    def test_init(self, default_env, wait_for_status):
        default_env.return_value = 'local'
        wait_for_status.return_value = mock_status

        sentry = Talisman(['meteor'])

        self.assertTrue('meteor/0' in sentry.unit)
        self.assertTrue('meteor/1' in sentry.unit)

    @patch.object(Talisman, 'wait_for_status')
    @patch('amulet.sentry.helpers.default_environment')
    def test_getitem(self, default_env, wait_for_status):
        default_env.return_value = 'local'
        wait_for_status.return_value = mock_status

        sentry = Talisman(['meteor'])

        self.assertEqual(sentry['meteor/0'], sentry.unit['meteor/0'])
        self.assertEqual(sentry['meteor'], list(sentry.unit.values()))




import pkgutil


modules = []
for i in [name for _, name, _ in pkgutil.iter_modules(['modules'])]:
    exec("from modules.%s import Module" % i)
    exec("modules.append(Module)")


API = type("API", tuple(modules), {})



"""Unit test for amulet.wait"""

import os
import unittest
import json
import yaml

from amulet import Deployment
from amulet.deployer import CharmCache
from mock import patch, MagicMock, call

RAW_ENVIRONMENTS_YAML = '''
default: gojuju
environments:
  gojuju:
    type: cloud
    access-key: xxx
    secret-key: yyy
    control-bucket: gojuju-xyxyz
    admin-secret: zyxyx
    default-series: world
  pyjuju:
    type: cloud
    access-key: xxx
    secret-key: yyy
    control-bucket: juju-xyxyz
    admin-secret: zyxyxy
    default-series: world
    juju-origin: ppa
    ssl-hostname-verification: true'''


class DeployerTests(unittest.TestCase):
    def test_init(self):
        d = Deployment(juju_env='gojuju')
        self.assertEqual('precise', d.series)
        self.assertEqual('gojuju', d.juju_env)
        self.assertEqual({}, d.services)
        self.assertEqual(True, d.use_sentries)
        self.assertEqual({}, d._sentries)
        self.assertEqual([], d.relations)
        d.cleanup()

    def test_load(self):
        d = Deployment(juju_env='gojuju')
        schema = '{"mybundle": {"series": "raring", "services": {"wordpress": \
                  {"branch": "lp:~charmers/charms/precise/wordpress/trunk"}, \
                  "mysql": {"options": {"tuning": "fastest"}, \
                  "branch": "lp:~charmers/charms/precise/mysql/trunk"}}, \
                  "relations": [["mysql:db", "wordpress:db"]]}}'
        dmap = json.loads(schema)
        with patch.object(d, 'add') as add:
            d.load(dmap)
        self.assertEqual(d.juju_env, 'gojuju')
        self.assertEqual(dmap['mybundle']['relations'], d.relations)
        self.assertEqual(dmap['mybundle']['series'], d.series)
        add.assert_has_calls([
            call('wordpress', charm=None, units=1),
            call('mysql', charm=None, units=1)],
            any_order=True
        )
        d.cleanup()

    @patch('amulet.deployer.get_charm')
    def test_add(self, mcharm):
        charm = mcharm.return_value
        charm.subordinate = False
        charm.code_source = {'location':
                             'lp:~charmers/charms/precise/charm/trunk'}
        charm.url = 'cs:precise/charm'

        d = Deployment(juju_env='gojuju')
        d.add('charm')
        self.assertEqual({'charm': {'charm': 'cs:precise/charm'}}, d.services)
        d.cleanup()

    @patch('amulet.deployer.get_charm')
    def test_add_branch(self, mcharm):
        charm = mcharm.return_value
        charm.subordinate = False
        charm.code_source = {'location': 'lp:~foo/charms/precise/baz/trunk'}
        charm.url = None

        d = Deployment(juju_env='gojuju')
        d.add('bar', 'cs:~foo/baz')
        self.assertEqual({'bar':
                          {'branch': 'lp:~foo/charms/precise/baz/trunk'}},
                         d.services)
        d.cleanup()

    @patch('amulet.deployer.get_charm')
    def test_add_units(self, mcharm):
        charm = mcharm.return_value
        charm.subordinate = False
        charm.code_source = {'location': 'lp:charms/charm'}
        charm.url = None

        d = Deployment(juju_env='gojuju')
        d.add('charm', units=2)
        self.assertEqual({'charm': {'branch': 'lp:charms/charm',
                                    'num_units': 2}}, d.services)
        d.cleanup()

    def _make_mock_status(self, d):
        def _mock_status(juju_env):
            status = dict(services={})
            total_units = 1
            for service in d.services:
                status['services'][service] = dict(units={})
                for unit in range(d.services[service].get('num_units', 1)):
                    total_units += 1
                    status['services'][service]['units'][
                        '{}/{}'.format(service, unit)] = {
                        'public-address': '10.0.3.{}'.format(total_units)}
            status['services']['relation-sentry'] = {
                'units': {
                    'relation-sentry/0': {
                        'public-address': '10.0.3.1'}}}
            return status
        return _mock_status

    @patch('amulet.helpers.environments')
    @patch('amulet.sentry.waiter.status')
    @patch('amulet.deployer.subprocess')
    @patch('amulet.deployer.get_charm')
    def test_add_unit(self, mcharm, subprocess, waiter_status, environments):
        charm = mcharm.return_value
        charm.subordinate = False
        charm.code_source = {'location': 'lp:charms/charm'}
        charm.url = None
        charm.series = 'precise'

        environments.return_value = yaml.safe_load(RAW_ENVIRONMENTS_YAML)

        d = Deployment(juju_env='gojuju')
        waiter_status.side_effect = self._make_mock_status(d)
        d.add('charm', units=1)
        d.setup()
        with patch('amulet.deployer.juju'):
            d.add_unit('charm')
        self.assertTrue('charm/1' in d.sentry.unit)

        d.cleanup()

    @patch('amulet.helpers.environments')
    @patch('amulet.sentry.waiter.status')
    @patch('amulet.deployer.subprocess')
    @patch('amulet.deployer.get_charm')
    def test_add_unit_error(self, mcharm, subprocess, waiter_status, environments):
        def mock_unit_error(f, service, unit_name):
            def _mock_unit_error(juju_env):
                status = f(juju_env)
                unit = status['services'][service]['units'].get(unit_name)
                if not unit:
                    return status
                unit['agent-state'] = 'error'
                unit['agent-state-info'] = 'hook failed: install'
                return status
            return _mock_unit_error

        charm = mcharm.return_value
        charm.subordinate = False
        charm.code_source = {'location': 'lp:charms/charm'}
        charm.url = None
        charm.series = 'precise'

        environments.return_value = yaml.safe_load(RAW_ENVIRONMENTS_YAML)

        d = Deployment(juju_env='gojuju')
        waiter_status.side_effect = mock_unit_error(
            self._make_mock_status(d), 'charm', 'charm/1')
        d.add('charm', units=1)
        d.setup()
        with patch('amulet.deployer.juju'):
            self.assertRaisesRegexp(
                Exception, 'Error on unit charm/1: hook failed: install',
                d.add_unit, 'charm')
        d.cleanup()

    @patch('amulet.deployer.get_charm')
    def test_add_error(self, mcharm):
        d = Deployment(juju_env='gojuju')
        d.add('bar')
        self.assertRaises(ValueError, d.add, 'bar')
        d.cleanup()

    @patch('amulet.deployer.get_charm')
    def test_relate(self, mcharm):
        a = mcharm.return_value
        a.provides = {'f': {'interface': 'test'}}
        a.requires = {'b': {'interface': 'test'}}

        d = Deployment(juju_env='gojuju')
        d.add('bar')
        d.add('foo')
        d.relate('foo:f', 'bar:b')
        self.assertEqual(d.relations, [['foo:f', 'bar:b']])
        d.cleanup()

    def test_relate_too_few(self):
        d = Deployment(juju_env='gojuju')
        self.assertRaises(LookupError, d.relate, 'foo:f')
        d.cleanup()

    def test_relate_no_relation(self):
        d = Deployment(juju_env='gojuju')
        self.assertRaises(ValueError, d.relate, 'foo', 'bar:a')
        d.cleanup()

    @patch('amulet.deployer.get_charm')
    def test_relate_relation_nonexist(self, mcharm):
        d = Deployment(juju_env='gojuju')
        d.add('bar')
        d.add('foo')

        self.assertRaises(ValueError, d.relate, 'foo:f', 'bar:b')
        d.cleanup()

    def test_relate_not_deployed(self):
        d = Deployment(juju_env='gojuju')
        self.assertRaises(ValueError, d.relate, 'foo:f', 'bar:a')
        d.cleanup()

    @patch('amulet.deployer.get_charm')
    def test_configure(self, mcharm):
        charm = mcharm.return_value
        charm.subordinate = False
        charm.code_source = {'location':
                             'lp:~charmers/charms/precise/wordpress/trunk'}
        charm.url = 'cs:precise/wordpress'

        d = Deployment(juju_env='gojuju')
        d.add('wordpress')
        d.configure('wordpress', {'tuning': 'optimized'})
        d.configure('wordpress', {'wp-content': 'f', 'port': 100})
        self.assertEqual({'wordpress':
                          {'charm': 'cs:precise/wordpress',
                           'options': {'tuning': 'optimized',
                                       'wp-content': 'f',
                                       'port': 100}}}, d.services)
        d.cleanup()

    def test_configure_not_deployed(self):
        d = Deployment(juju_env='gojuju')
        self.assertRaises(ValueError, d.configure, 'wordpress',
                          {'tuning': 'optimized'})
        d.cleanup()

    @patch('amulet.deployer.get_charm')
    def test_expose(self, mcharm):
        charm = mcharm.return_value
        charm.subordinate = False
        charm.code_source = {'location':
                             'lp:~charmers/charms/precise/wordpress/trunk'}
        charm.url = None

        d = Deployment(juju_env='gojuju')
        d.add('wordpress')
        d.expose('wordpress')
        self.assertEqual({'wordpress':
                          {'branch':
                           'lp:~charmers/charms/precise/wordpress/trunk',
                           'expose': True}}, d.services)
        d.cleanup()

    def test_expose_not_deployed(self):
        d = Deployment(juju_env='gojuju')
        self.assertRaises(ValueError, d.expose, 'wordpress')
        d.cleanup()

    @patch('amulet.deployer.get_charm')
    def test_schema(self, mcharm):
        wpmock = MagicMock()
        mysqlmock = MagicMock()
        wpmock.subordinate = False
        wpmock.code_source = {'location':
                              'lp:~charmers/charms/precise/wordpress/trunk'}
        wpmock.requires = {'db': {'interface': 'mysql'}}
        wpmock.url = None
        mysqlmock.subordinate = False
        mysqlmock.code_source = {'location':
                                 'lp:~charmers/charms/precise/mysql/trunk'}
        mysqlmock.provides = {'db': {'interface': 'mysql'}}
        mysqlmock.url = None

        mcharm.side_effect = [mysqlmock, wpmock]
        d = Deployment(juju_env='gojuju', sentries=False)
        d.add('mysql')
        d.configure('mysql', {'tuning': 'fastest'})
        d.add('wordpress')
        d.relate('mysql:db', 'wordpress:db')
        schema = {'gojuju': {'services': {'mysql': {
            'branch': 'lp:~charmers/charms/precise/mysql/trunk',
            'options': {'tuning': 'fastest'}},
            'wordpress': {'branch':
                          'lp:~charmers/charms/precise/wordpress/trunk'}},
            'series': 'precise', 'relations': [['mysql:db', 'wordpress:db']]}}
        self.assertEqual(schema, d.schema())
        d.cleanup()

    def test_build_sentries_writes_relationship_sentry_metadata(self):
        """Even if there are no relations the metadata.yaml is written."""
        d = Deployment(juju_env='gojuju', sentries=True)
        d.build_sentries()
        self.assertIn('metadata.yaml',
                      os.listdir(d.relationship_sentry.charm))
        d.cleanup()

    @patch.dict('os.environ', {'JUJU_TEST_CHARM': 'charmbook'})
    def test_juju_test_charm(self):
        d = Deployment(juju_env='gogo')
        self.assertEqual('charmbook', d.charm_name)
        d.cleanup()

    def test_add_post_deploy(self):
        d = Deployment(juju_env='gogo')
        d.deployed = True
        self.assertRaises(NotImplementedError, d.add, 'mysql')
        d.cleanup()

    def test_relate_post_deploy(self):
        d = Deployment(juju_env='gogo')
        d.deployed = True
        self.assertRaises(NotImplementedError, d.relate, 'mysql:db',
                          'wordpress:db')
        d.cleanup()

    def test_unrelate_not_enough(self):
        d = Deployment(juju_env='gogo')
        self.assertRaises(LookupError, d.unrelate, 'mysql')
        d.cleanup()

    def test_unrelate_too_many(self):
        d = Deployment(juju_env='gogo')
        self.assertRaises(LookupError, d.unrelate,
                          'mysql', 'wordpress', 'charm')
        d.cleanup()

    def test_unrelate_bad_service_format(self):
        d = Deployment(juju_env='gogo')
        d.deployed = True
        with self.assertRaises(ValueError) as e:
            d.unrelate('mysql', 'wordpress')
            self.assertEqual(
                str(e), 'All relations must be explicit, service:relation')
        d.cleanup()

    @patch.object(Deployment, '_get_sentry_relations')
    @patch('amulet.deployer.juju')
    def test_unrelate(self, mj, _get_sentry_relations):
        _get_sentry_relations.return_value = [
            ['relation-sentry:provides-mysql_db-charm_db', 'mysql:db'],
            ['relation-sentry:requires-mysql_db-charm_db', 'charm:db'],
        ]

        d = Deployment(juju_env='gogo')
        d.deployed = True
        d.unrelate('mysql:db', 'charm:db')
        mj.assert_has_calls([
            call(['remove-relation',
                  'relation-sentry:provides-mysql_db-charm_db', 'mysql:db']),
            call(['remove-relation',
                  'relation-sentry:requires-mysql_db-charm_db', 'charm:db']),
            ])
        d.cleanup()

    def test_unrelate_post_deploy(self):
        d = Deployment(juju_env='gogo')
        self.assertRaises(NotImplementedError, d.unrelate, 'mysql:db',
                          'wordpress:db')
        d.cleanup()

    def test_get_sentry_relations(self):
        d = Deployment(juju_env='gogo')
        d.relations = [
            ['relation-sentry:provides-mysql_db-charm_db', 'mysql:db'],
            ['relation-sentry:requires-mysql_db-charm_db', 'charm:db'],
        ]
        self.assertEqual(
            d._get_sentry_relations('charm:db', 'mysql:db'),
            d.relations)
        self.assertEqual(
            d._get_sentry_relations('mysql:db', 'charm:db'),
            list(reversed(d.relations)))
        d.cleanup()

    def test_get_sentry_relations_not_found(self):
        d = Deployment(juju_env='gogo')
        d.relations = []
        with self.assertRaises(LookupError) as e:
            d._get_sentry_relations('charm:db', 'mysql:db')
            self.assertEqual(
                str(e),
                'Could not find relation between charm:db and mysql:db')
        d.cleanup()

    @patch('amulet.deployer.juju')
    def test_remove_unit(self, mj):
        d = Deployment(juju_env='gogo')
        d.add('mysql', units=2)
        d.deployed = True
        d.remove_unit('mysql/1')
        mj.assert_called_with(['remove-unit', 'mysql/1'])
        d.cleanup()

    def test_remove_unit_env_not_setup(self):
        d = Deployment(juju_env='gogo')
        d.add('mysql', units=2)
        self.assertRaises(NotImplementedError, d.remove_unit, 'mysql/0')
        d.cleanup()

    def test_remove_unit_no_args(self):
        d = Deployment(juju_env='gogo')
        d.add('mysql', units=2)
        d.deployed = True
        self.assertRaises(ValueError, d.remove_unit)
        d.cleanup()

    def test_remove_unit_not_a_unit(self):
        d = Deployment(juju_env='gogo')
        d.add('mysql', units=2)
        d.deployed = True
        self.assertRaises(ValueError, d.remove_unit, 'mysql/1', 'lolk')
        d.cleanup()

    def test_remove_unit_not_deployed(self):
        d = Deployment(juju_env='gogo')
        d.add('mysql', units=2)
        d.deployed = True
        self.assertRaises(ValueError, d.remove_unit, 'wordpress/1')
        d.cleanup()

    def test_setup(self):
        pass


class CharmCacheTest(unittest.TestCase):
    def test_init(self):
        c = CharmCache('mytestcharm')
        self.assertEqual(c.test_charm, 'mytestcharm')

    @patch('amulet.deployer.get_charm')
    def test_getitem_service(self, get_charm):
        c = CharmCache('mytestcharm')
        charm = c['myservice']
        self.assertEqual(charm, get_charm.return_value)
        get_charm.assert_called_once_with('myservice')

        get_charm.reset_mock()
        charm2 = c['myservice']
        self.assertEqual(charm, charm2)
        self.assertFalse(get_charm.called)

    @patch('amulet.deployer.get_charm')
    def test_getitem_testcharm(self, get_charm):
        c = CharmCache('mytestcharm')
        charm = c['mytestcharm']
        self.assertEqual(charm, get_charm.return_value)
        get_charm.assert_called_once_with(os.getcwd())

    @patch('amulet.deployer.get_charm')
    def test_fetch_service(self, get_charm):
        c = CharmCache('mytestcharm')
        charm = c.fetch('myservice')
        self.assertEqual(charm, get_charm.return_value)
        get_charm.assert_called_once_with('myservice')

        get_charm.reset_mock()
        charm2 = c['myservice']
        self.assertEqual(charm, charm2)
        self.assertFalse(get_charm.called)

    @patch('amulet.deployer.get_charm')
    def test_fetch_charm(self, get_charm):
        c = CharmCache('mytestcharm')
        charm = c.fetch('myservice', 'anothercharm')
        self.assertEqual(charm, get_charm.return_value)
        get_charm.assert_called_once_with('anothercharm')

    @patch('amulet.deployer.get_charm')
    def test_fetch_testcharm(self, get_charm):
        c = CharmCache('mytestcharm')
        charm = c.fetch('myservice', 'mytestcharm')
        self.assertEqual(charm, get_charm.return_value)
        get_charm.assert_called_once_with(os.getcwd())




import os
import sys
import tempfile
import unittest
import yaml
import shutil

from mock import patch, Mock, call
from amulet.deployer import _default_sentry_template

from amulet.charm import (
    Builder,
    run_bzr,
    get_relation,
    Charm,
    LocalCharm,
    LaunchpadCharm,
    setup_bzr,
)


class BuilderTest(unittest.TestCase):
    def test_does_not_create_yaml_tags(self):
        """Instead of creating yaml safe_load will refuse, fail at write"""
        class customstr(str):
            """A custom Python type yaml would serialise tagged"""
        self.assertIn("!!", yaml.dump(customstr("a")))
        builder = Builder(customstr("acharm"), _default_sentry_template)
        self.assertRaises(yaml.representer.RepresenterError,
                          builder.write_metadata)
        shutil.rmtree(builder.charm)


class RunBzrTest(unittest.TestCase):
    def test_run_bzr(self):
        out = run_bzr(["rocks"], ".")
        self.assertEquals(out, "It sure does!\n")

    @patch('subprocess.Popen')
    def test_run_bzr_traceback(self, mp):
        mp.side_effect = [Exception("AssertionError: always fails")]
        self.assertRaisesRegexp(Exception, "AssertionError: always fails",
                                run_bzr, ["assert-fail"], ".")

    @patch('subprocess.Popen')
    def test_run_bzr_oserror(self, mp):
        mp.side_effect = [OSError(1, "Command failed")]
        self.assertRaisesRegexp(OSError, "Command failed",
                                run_bzr, ["assert-fail"], ".")

    def test_run_bzr_missing(self):
        env = os.environ.copy()
        env["PATH"] = ""
        self.assertRaisesRegexp(OSError, "bzr not found, do you have Bazaar "
                                "installed?", run_bzr, ["version"], ".",
                                env=env)


class SetupBzrTest(unittest.TestCase):
    @patch('amulet.helpers.run_bzr')
    def test_setup_bzr(self, mp):
        mp.side_effect = [IOError("bzr command failed!"), None, None]
        setup_bzr('/path')
        self.assertEqual(mp.call_args_list,
                         [call(['whoami'], '/path'),
                          call(['whoami', 'amulet@dummy-user.tld'], '/path'),
                          call(['init'], '/path')])


RAW_METADATA_YAML = '''
name: charm-name
description: Whatever man
requires:
  relation:
    interface: iname
provides:
  plation:
    interface: aniname
'''


class GetRelationTest(unittest.TestCase):
    @patch('os.path.exists')
    @patch('builtins.open' if sys.version_info > (3,) else '__builtin__.open')
    def test_get_relation_local_charm(self, mock_open, mexists):
        mock_open.return_value.__enter__ = lambda s: s
        mock_open.return_value.__exit__ = Mock()
        mock_open.return_value.read.return_value = RAW_METADATA_YAML
        mexists.return_value = True
        self.assertEqual(('provides', 'aniname'),
                         get_relation('/path/to/charm', 'plation'))

    @patch('amulet.charm.Charm')
    def test_get_relation_remote(self, mcharm):
        cdata = {'charm': {'relations': {'requires': {'relname': {'interface':
                                                                  'iname'}}}}}
        mcharm.return_value = Charm.from_charmdata(cdata)
        self.assertEqual(('requires', 'iname'), get_relation('c', 'relname'))

    @patch('os.path.exists')
    @patch('builtins.open' if sys.version_info > (3,) else '__builtin__.open')
    def test_no_relations(self, mock_open, mexists):
        BAD_METADATA_YAML = '''
        name: charm-name
        description: Whatever man
        '''
        mock_open.return_value.__enter__ = lambda s: s
        mock_open.return_value.__exit__ = Mock()
        mock_open.return_value.read.return_value = BAD_METADATA_YAML
        mexists.return_value = True
        self.assertRaises(Exception, get_relation, '/path/2/bad/charm', 'noop')

    @patch('os.path.exists')
    @patch('builtins.open' if sys.version_info > (3,) else '__builtin__.open')
    def test_no_match(self, mock_open, mexists):
        mock_open.return_value.__enter__ = lambda s: s
        mock_open.return_value.__exit__ = Mock()
        mock_open.return_value.read.return_value = RAW_METADATA_YAML
        mexists.return_value = True
        self.assertEqual((None, None), get_relation('/path/charm', 'noop'))


class LocalCharmTest(unittest.TestCase):
    @patch('os.path.exists')
    @patch.object(LocalCharm, '_load')
    def test_parse(self, mlc_l, me):
        me.return_value = True
        mlc_l.return_value = {'name': 'test',
                              'author': 'ohkay',
                              'provides': {'rel': {'interface': 'test'}}}
        c = LocalCharm('/path/to/test')
        self.assertEqual('test', c.name)
        mlc_l.assert_called_with('/path/to/test/metadata.yaml')

    def test_make_temp(self):
        charm_dir = tempfile.mkdtemp()
        metadata = {}
        with open(os.path.join(charm_dir, 'metadata.yaml'), 'w') as f:
            f.write(yaml.dump(metadata))

        c = LocalCharm(charm_dir)
        code_source = c.code_source['location']
        self.assertTrue(code_source != charm_dir)
        self.assertTrue(code_source.startswith('/tmp/charm'))
        self.assertTrue(os.path.exists(os.path.join(code_source, '.bzr')))
        self.assertTrue(os.path.exists(os.path.join(
            code_source, 'metadata.yaml')))

        del c
        self.assertFalse(os.path.exists(os.path.join(code_source, '../')))


class LaunchpadCharmTest(unittest.TestCase):
    pass




import cherrypy
import json_rpc
import subprocess


class Module (object):

    # Expose secure when we get admin-secret working
    @json_rpc.expose_anonymous
    def run(self):
        if not cherrypy.request.method in ['POST', 'PUT']:
            raise cherrypy.HTTPError(405)

        command = cherrypy.request.body.read().decode()
        try:
            code = 0
            results = subprocess.check_output(command, shell=True)
        except subprocess.CalledProcessError as e:
            cherrypy.response.status = 500
            code = e.returncode
            results = e.output

        return {'code': code, 'output': results.decode().strip()}



import argparse
import inspect

import amulet


def setup_parser():
    parser = argparse.ArgumentParser(prog='amulet')
    subparsers = parser.add_subparsers(help='sub-command help')

    for name, obj in inspect.getmembers(amulet):
        if inspect.ismodule(obj):
            for baby_name, baby_obj in inspect.getmembers(obj):
                if baby_name == 'setup_parser':
                    getattr(obj, baby_name)(subparsers)

    return parser


def main():
    p = setup_parser()
    args = p.parse_args()
    args.func(args)

if __name__ == '__main__':
    main()



#!/usr/bin/python

import os
import sys
import json
import glob
import subprocess


relation_dir = '/opt/relations'
unit_number = os.environ['JUJU_REMOTE_UNIT'].split('/')[1]
data_dir = os.path.join(relation_dir, os.environ['JUJU_RELATION'], unit_number)
endpoint, relation = os.environ['JUJU_RELATION'].split('-', 1)

if not os.path.exists(data_dir):
    os.makedirs(data_dir)

with open(os.path.join(data_dir, 'id'), 'w') as f:
    f.write(os.environ['JUJU_RELATION_ID'])

with open(os.path.join(data_dir, 'data'), 'w') as f:
    try:
        data = subprocess.check_output(['relation-get', '--format=json'])
    except:
        pass
    else:
        f.write(data)

if endpoint == "requires":
    otherend = "provides"
else:
    otherend = "requires"

other_data_dir = os.path.join(relation_dir, "%s-%s" % (otherend, relation))
subprocess.call(['juju-log', 'Running: %s %s' % (os.environ['JUJU_RELATION'],
                                                 otherend)])

if not os.path.exists(other_data_dir):
    sys.exit(0)

for unit_data_dir in glob.glob(os.path.join(other_data_dir, '*')):
    unit = os.path.basename(unit_data_dir)
    with open(os.path.join(unit_data_dir, 'id')) as f:
        relation_id = f.read().strip()

    with open(os.path.join(unit_data_dir, 'data')) as f:
        relation_data = json.loads(f.read().strip())

    subprocess.call(['juju-log', json.dumps(relation_data)])
    relation_cmd_line = ['relation-set']
    for key, val in relation_data.items():
        if val is None:
            relation_cmd_line.append('{}='.format(key))
        else:
            relation_cmd_line.append('{}={}'.format(key, val))

    subprocess.call(['juju-log', ' '.join(relation_cmd_line)])
    subprocess.check_call(relation_cmd_line)

    # Send our data to other units
    if data:
        rel_cmd = ['relation-set', '-r', relation_id]
        for key, val in json.loads(data).items():
            if val is None:
                rel_cmd.append('{}='.format(key))
            else:
                rel_cmd.append('{}={}'.format(key, val))

            subprocess.call(['juju-log', ' '.join(rel_cmd)])
            subprocess.check_call(rel_cmd)




import os
import json_rpc

PROC_DIR = '/proc'


class Module (object):
    @json_rpc.expose_anonymous
    def juju(self):
        pids = [pid for pid in os.listdir(PROC_DIR) if pid.isdigit()]

        hook = None
        for pid in pids:
            with open(os.path.join(PROC_DIR, pid, 'cmdline'), 'r') as p:
                cmd = p.read()
                if '/var/lib/juju/agents/' in cmd:
                    hook = os.path.basename(cmd)

        if hook:
            return {'hook': hook}
        else:
            return {}



from setuptools import setup

install_requires = [
    'requests',
    'charmworldlib',
    'PyYAML'
]

tests_require = [
    'coverage',
    'nose',
    'pep8',
]


setup(
    name='amulet',
    version='1.6.0',
    description='Tools to help with writing Juju Charm Functional tests',
    install_requires=install_requires,
    package_data={'amulet': ['charms/sentry/hooks/*',
                             'charms/sentry/src/*.*',
                             'charms/sentry/src/*/*']},
    author='Marco Ceppi',
    author_email='marco@ceppi.net',
    url="https://launchpad.net/amulet",
    packages=['amulet'],
    entry_points={
        'console_scripts': [
            'amulet=amulet.cli:main'
        ]
    }
)




import os
import json
import cherrypy
import json_rpc


class Module (object):
    @json_rpc.expose_anonymous_dont_jsonify
    def file(self, action=None, name=None):
        if not cherrypy.request.method in ['GET']:
            raise cherrypy.HTTPError(405)

        if not name:
            raise cherrypy.HTTPError(400)

        if action and action != 'contents':
            raise cherrypy.HTTPError(400)

        if not os.path.isfile(name):
            raise cherrypy.HTTPError(404)

        if not action:
            return json.dumps(self.fs_data(name), indent=4)

        with open(name, 'r') as f:
            contents = f.read()

        return contents

    @json_rpc.expose_anonymous
    def directory(self, action=None, name=None):
        if not cherrypy.request.method in ['GET']:
            raise cherrypy.HTTPError(405)

        if not name:
            raise cherrypy.HTTPError(400)

        if action and action != 'contents':
            raise cherrypy.HTTPError(400)

        if not os.path.isdir(name):
            raise cherrypy.HTTPError(404)

        if not action:
            return self.fs_data(name)

        contents = {'files': [], 'directories': []}
        for fd in os.listdir(name):
            if os.path.isfile('%s/%s' % (name, fd)):
                contents['files'].append(fd)
            else:
                contents['directories'].append(fd)

        return contents

    def fs_data(self, path):
        fs_stat = os.stat(path)

        return {'mtime': fs_stat.st_mtime,
                'size': fs_stat.st_size,
                'uid': fs_stat.st_uid,
                'gid': fs_stat.st_gid,
                'mode': oct(fs_stat.st_mode)}




import requests

from . import waiter
from . import helpers
from .charm import get_relation

try:
    from urllib.parse import urlencode
except ImportError:
    from urllib import urlencode


class SentryError(Exception):
    pass


class Sentry(object):
    def __init__(self, address, port=9001):
        self.config = {}
        self.config['address'] = 'https://%s:%s' % (address, port)

    def file(self, filename):
        return self.file_stat(filename)

    def file_stat(self, filename):
        raise NotImplemented()

    def file_contents(self, filename):
        raise NotImplemented()

    def directory(self, *args):
        return self.directory_stat(*args)

    def directory_stat(self, *args):
        raise NotImplemented()

    def directory_contents(self, *args):
        return self.directory_listing(*args)

    def directory_listing(self, *args):
        raise NotImplemented()

    def juju_agent(self):
        try:
            return self.query('/juju').json()
        except:
            return None

    def query(self, endpoint, query={}, data=None):
        return self._fetch(self.config['address'], endpoint, query, data)

    def _fetch(self, address, endpoint, query={}, data=None):
        url = "%s/%s?%s" % (address, endpoint, urlencode(query))
        if data:
            return requests.post(url, data=data, verify=False)
        else:
            return requests.get(url, verify=False)


class UnitSentry(Sentry):
    @classmethod
    def fromunit(cls, unit):
        pass

    @classmethod
    def fromunitdata(cls, unit, unit_data, sentry, port=9001):
        address = unit_data['public-address']
        unitsentry = cls(address)
        unitsentry.info = unit_data
        unitsentry.info['service'] = unit.split('/')[0]
        unitsentry.info['unit'] = unit.split('/')[1]
        unitsentry.config['sentry'] = 'https://%s:%s' % (sentry, port)
        return unitsentry

    def file_stat(self, filename):
        r = self._fetch_filesystem('/file', {'name': filename})
        return r.json()

    def _fetch_filesystem(self, endpoint, params):
        r = self.query(endpoint, params)
        if r.status_code == 404:
            raise IOError('%s does not exist on unit' % params['name'])
        elif r.status_code != 200:
            raise SentryError('API returned the following: %s' % r.status_code)

        return r

    def file_contents(self, filename):
        r = self._fetch_filesystem('/file/contents', {'name': filename})
        return r.text

    def directory_stat(self, path):
        r = self._fetch_filesystem('/directory', {'name': path})
        return r.json()

    def directory_listing(self, path):
        r = self._fetch_filesystem('/directory/contents', {'name': path})
        return r.json()

    def run(self, command):
        r = self.query('/run', data=command)
        results = r.json()
        return results['output'], results['code']

    #d.sentry.unit[].relation('db', 'mysql:db')
    def relation(self, from_rel, to_rel):
        # Build possible mappings, find the map, produce results
        potential_rel = ['%s:%s' % (self.info['service'], from_rel), to_rel]
        rel_data = get_relation(self.info['service'], from_rel)
        relations = ['-'.join([rel_data[0],
                     "-".join(potential_rel).replace(':', '_')])]
        potential_rel.reverse()
        relations.append('-'.join([rel_data[0],
                         "-".join(potential_rel).replace(':', '_')]))
        for relation in relations:
            r = self._fetch(self.config['sentry'],
                            '/relation/%s/%s' % (relation, self.info['unit']))
            if r.status_code == 200:
                return r.json()

        raise Exception('Relationship not found')


# Possibly use to build out instead of having code in setup()?
class Talisman(object):
    def __init__(self, services, rel_sentry='relation-sentry', juju_env=None):
        self.unit = {}
        self.service = {}

        if not juju_env:
            juju_env = helpers.default_environment()

        status = self.wait_for_status(juju_env, services)
        if rel_sentry in status['services']:
            rel_sentry_units = status['services'][rel_sentry]['units']
            rel_sentry_unit = rel_sentry_units['/'.join([rel_sentry, '0'])]
            rel_sentry_addr = rel_sentry_unit['public-address']
        else:
            raise Exception('No relationship sentry found')

        for service in services:
            if not service in status['services']:
                continue  # Raise something?

            # self.sentry.service[service] = ServiceSentry()

            service_status = status['services'][service]

            if not 'units' in service_status:
                continue  # It's a subordinate

            for unit in service_status['units']:
                unit_data = service_status['units'][unit]
                self.unit[unit] = UnitSentry.fromunitdata(unit, unit_data,
                                                          rel_sentry_addr)

    def __getitem__(self, service):
        """Return the UnitSentry object(s) for ``service``

        :param service: A string in one of two forms::
            "service_name"
            "service_name/unit_num"

        If the first form is used, a list (possibly empty) of all the
        UnitSentry objects for that service is returned.

        If the second form is used, a single object, the UnitSentry for
        the specified unit, is returned.

        Examples::

            >>> d
            <amulet.deployer.Deployment object at 0x7fac83ce8438>
            >>> d.schema()['local']['services']['meteor']['num_units']
            2
            >>> # get UnitSentry for specific unit
            >>> meteor_0 = d.sentry['meteor/0']
            >>> print(meteor_0.info['service'], meteor_0.info['unit'])
            meteor 0
            >>> # get all UnitSentry objects for a service
            >>> for sentry in d.sentry['meteor']:
            ...     print(sentry.info['service'], sentry.info['unit'])
            ...
            meteor 1
            meteor 0
            >>>

        """
        single_unit = '/' in service

        def match(service, unit_name):
            if single_unit:
                return service == unit_name
            return service == unit_name.split('/')[0]

        unit_sentries = [unit_sentry
                         for unit_name, unit_sentry in self.unit.items()
                         if match(service, unit_name)]

        if single_unit and unit_sentries:
            return unit_sentries[0]

        return unit_sentries

    def wait_for_status(self, juju_env, services, timeout=300):
        """Return env status, but only after all units have a
        public-address assigned.

        Raises if a unit reaches error state, or if public-address not
        available for all units before timeout expires.

        """
        try:
            with helpers.timeout(timeout):
                while True:
                    ready = True
                    status = waiter.status(juju_env)
                    for service in services:
                        if not 'units' in status['services'][service]:
                            continue
                        for unit, unit_dict in \
                                status['services'][service]['units'].items():
                            if 'error' == unit_dict.get('agent-state'):
                                raise Exception('Error on unit {}: {}'.format(
                                    unit, unit_dict.get('agent-state-info')))
                            if 'public-address' not in unit_dict:
                                ready = False
                    if ready:
                        return status
        except helpers.TimeoutError:
            raise helpers.TimeoutError(
                'public-address not set for'
                'all units after {}s'.format(timeout))
        except:
            raise

    def wait(self, timeout=300):
        #for unit in self.unit:
        ready = False
        try:
            with helpers.timeout(timeout):
                # Make sure we're in a 'started' state across the board
                waiter.wait(timeout=timeout)
                while not ready:
                    for unit in self.unit.keys():
                        status = self.unit[unit].juju_agent()
                        # Check if we have a hook key and it's not None
                        if status is None:
                            ready = False
                            break
                        if 'hook' in status and status['hook']:
                            ready = False
                            break
                        else:
                            ready = True
        except:
            raise

    def _sync(self):
        pass


class ServiceSentry(Sentry):
    pass



import sys
import base64
import hashlib
import json
import os
import shutil
import tempfile

from mock import patch
from unittest import TestCase

server_location = os.path.join(
    os.path.dirname(__file__), '../', 'amulet', 'charms', 'sentry', 'src')
sys.path.insert(0, server_location)

os.chdir(server_location)

import cherrypy

import server
import json_rpc
from modules import docs
from modules import filesystem
from modules import juju
from modules import relations
from modules import run


class TestInstance(TestCase):
    def setUp(self):
        super(TestInstance, self).setUp()

        users = {'admin': hashlib.sha512(u'admin'.encode()).hexdigest()}
        self.tempdir = tempfile.mkdtemp()
        os.chdir(self.tempdir)
        with open('users.json', 'w') as f:
            json.dump(users, f)

    def tearDown(self):
        shutil.rmtree(self.tempdir)
        os.chdir(server_location)

    def test_check_credentials(self):
        inst = server.Instance()

        self.assertTrue(inst._check_credentials(u'admin', u'admin'))
        self.assertFalse(inst._check_credentials(u'admin', u'foo'))

    def test_validate(self):
        inst = server.Instance()

        basic_auth_header = u'Basic {}'.format(
            base64.b64encode(b'admin:admin').decode())
        self.assertFalse(inst._validate("Digest"))
        self.assertTrue(inst._validate(basic_auth_header))


class TestDocs(TestCase):
    def test_index(self):
        mod = docs.Module()

        self.assertEqual(mod.index().read(), u"You shouldn't be here.\n")


class TestFilesystem(TestCase):
    @patch('modules.filesystem.cherrypy.request')
    def test_file_not_GET(self, request):
        request.method = 'POST'

        mod = filesystem.Module()

        with self.assertRaises(cherrypy.HTTPError) as e:
            mod.file()
            self.assertEqual(e.status, 405)

    @patch('modules.filesystem.cherrypy.request')
    def test_file_no_name(self, request):
        request.method = 'GET'

        mod = filesystem.Module()

        with self.assertRaises(cherrypy.HTTPError) as e:
            mod.file()
            self.assertEqual(e.status, 400)

    @patch('modules.filesystem.cherrypy.request')
    def test_file_bad_action(self, request):
        request.method = 'GET'

        mod = filesystem.Module()

        with self.assertRaises(cherrypy.HTTPError) as e:
            mod.file(action='list')
            self.assertEqual(e.status, 400)

    @patch('modules.filesystem.cherrypy.request')
    def test_file_no_such_file(self, request):
        request.method = 'GET'

        mod = filesystem.Module()

        with self.assertRaises(cherrypy.HTTPError) as e:
            mod.file(name='/tmp/no-such-file')
            self.assertEqual(e.status, 404)

    @patch('modules.filesystem.cherrypy.request')
    def test_file_stat(self, request):
        request.method = 'GET'

        mod = filesystem.Module()

        with tempfile.NamedTemporaryFile() as f:
            r = json.loads(mod.file(name=f.name).decode())
            stat = os.stat(f.name)
            self.assertEqual(r['mtime'], stat.st_mtime)
            self.assertEqual(r['size'], stat.st_size)
            self.assertEqual(r['uid'], stat.st_uid)
            self.assertEqual(r['gid'], stat.st_gid)
            self.assertEqual(r['mode'], oct(stat.st_mode))

    @patch('modules.filesystem.cherrypy.request')
    def test_file_contents(self, request):
        request.method = 'GET'

        mod = filesystem.Module()

        with tempfile.NamedTemporaryFile() as f:
            f.write(b'Test')
            f.flush()
            self.assertEqual(b'Test', mod.file(action='contents', name=f.name))

    @patch('modules.filesystem.cherrypy.request')
    def test_directory_not_GET(self, request):
        request.method = 'POST'

        mod = filesystem.Module()

        with self.assertRaises(cherrypy.HTTPError) as e:
            mod.directory()
            self.assertEqual(e.status, 405)

    @patch('modules.filesystem.cherrypy.request')
    def test_directory_no_name(self, request):
        request.method = 'GET'

        mod = filesystem.Module()

        with self.assertRaises(cherrypy.HTTPError) as e:
            mod.directory()
            self.assertEqual(e.status, 400)

    @patch('modules.filesystem.cherrypy.request')
    def test_directory_bad_action(self, request):
        request.method = 'GET'

        mod = filesystem.Module()

        with self.assertRaises(cherrypy.HTTPError) as e:
            mod.directory(action='list')
            self.assertEqual(e.status, 400)

    @patch('modules.filesystem.cherrypy.request')
    def test_directory_no_such_directory(self, request):
        request.method = 'GET'

        mod = filesystem.Module()

        with self.assertRaises(cherrypy.HTTPError) as e:
            mod.directory(name='/tmp/no-such-directory')
            self.assertEqual(e.status, 404)

    @patch('modules.filesystem.cherrypy.request')
    def test_directory_stat(self, request):
        request.method = 'GET'

        mod = filesystem.Module()

        tempdir = tempfile.mkdtemp()
        self.addCleanup(lambda: shutil.rmtree(tempdir))

        r = json.loads(mod.directory(name=tempdir).decode())
        stat = os.stat(tempdir)
        self.assertEqual(r['mtime'], stat.st_mtime)
        self.assertEqual(r['size'], stat.st_size)
        self.assertEqual(r['uid'], stat.st_uid)
        self.assertEqual(r['gid'], stat.st_gid)
        self.assertEqual(r['mode'], oct(stat.st_mode))

    @patch('modules.filesystem.cherrypy.request')
    def test_directory_contents(self, request):
        request.method = 'GET'

        mod = filesystem.Module()

        tempdir = tempfile.mkdtemp()
        tempfile.mkdtemp(prefix='mydir', dir=tempdir)
        tempfile.mkstemp(prefix='myfile', dir=tempdir)
        self.addCleanup(lambda: shutil.rmtree(tempdir))

        contents = json.loads(
            mod.directory(action='contents', name=tempdir).decode())
        self.assertEqual(len(contents['directories']), 1)
        self.assertEqual(len(contents['files']), 1)
        self.assertTrue(contents['directories'][0].startswith('mydir'))
        self.assertTrue(contents['files'][0].startswith('myfile'))


class TestJuju(TestCase):
    def test_juju_active_hook(self):
        tempdir = tempfile.mkdtemp()
        self.addCleanup(lambda: shutil.rmtree(tempdir))

        with patch('modules.juju.PROC_DIR', tempdir):
            os.mkdir(os.path.join(tempdir, '123'))
            with open(os.path.join(tempdir, '123', 'cmdline'), 'w') as f:
                f.write(u'/var/lib/juju/agents/charm/hooks/install')
                f.flush()

            mod = juju.Module()
            r = json.loads(mod.juju().decode())
            self.assertEqual(r['hook'], u'install')

    def test_juju_no_active_hook(self):
        tempdir = tempfile.mkdtemp()
        self.addCleanup(lambda: shutil.rmtree(tempdir))

        with patch('modules.juju.PROC_DIR', tempdir):
            os.mkdir(os.path.join(tempdir, '1'))
            with open(os.path.join(tempdir, '1', 'cmdline'), 'w') as f:
                f.write(u'/sbin/init')
                f.flush()

            mod = juju.Module()
            r = json.loads(mod.juju().decode())
            self.assertEqual(r, {})


class TestRelations(TestCase):
    def setUp(self):
        self.tempdir = tempfile.mkdtemp()
        patcher = patch('modules.relations.RELATIONS_DIR', self.tempdir)
        patcher.start()
        self.addCleanup(lambda: shutil.rmtree(self.tempdir))
        self.addCleanup(patcher.stop)

    @patch('modules.relations.cherrypy.request')
    def test_relations_not_GET(self, request):
        request.method = 'POST'

        mod = relations.Module()

        with self.assertRaises(cherrypy.HTTPError) as e:
            mod.relations()
            self.assertEqual(e.status, 405)

    @patch('modules.relations.cherrypy.request')
    def test_relations_no_dir(self, request):
        request.method = 'GET'

        mod = relations.Module()

        with patch('modules.relations.RELATIONS_DIR', '/tmp/no-such-dir'):
            self.assertEqual({}, json.loads(mod.relations().decode()))

    @patch('modules.relations.cherrypy.request')
    def test_relations(self, request):
        request.method = 'GET'

        mod = relations.Module()

        os.mkdir(os.path.join(self.tempdir, 'myrelation'))
        os.mkdir(os.path.join(self.tempdir, 'myrelation', '1'))

        expected = {'myrelation': ['1']}
        self.assertEqual(expected, json.loads(mod.relations().decode()))

    @patch('modules.relations.cherrypy.request')
    def test_relation_not_GET(self, request):
        request.method = 'POST'

        mod = relations.Module()

        with self.assertRaises(cherrypy.HTTPError) as e:
            mod.relation()
            self.assertEqual(e.status, 405)

    @patch('modules.relations.cherrypy.request')
    def test_relation_no_relation(self, request):
        request.method = 'GET'

        mod = relations.Module()

        with self.assertRaises(cherrypy.HTTPError) as e:
            mod.relation()
            self.assertEqual(e.status, 400)

    @patch('modules.relations.cherrypy.request')
    def test_relation_no_dir(self, request):
        request.method = 'GET'

        mod = relations.Module()

        with self.assertRaises(cherrypy.HTTPError) as e:
            mod.relation(relation='myrelation')
            self.assertEqual(e.status, 404)

    @patch('modules.relations.cherrypy.request')
    def test_relation_no_unit(self, request):
        request.method = 'GET'

        mod = relations.Module()

        os.mkdir(os.path.join(self.tempdir, 'myrelation'))
        os.mkdir(os.path.join(self.tempdir, 'myrelation', '1'))
        with open(os.path.join(
                self.tempdir, 'myrelation', '1', 'data'), 'w') as f:
            json.dump({'unit_data_key': 'unit_data_value'}, f)

        expected = {'1': {'unit_data_key': 'unit_data_value'}}
        self.assertEqual(
            expected, json.loads(mod.relation(relation='myrelation').decode()))

    @patch('modules.relations.cherrypy.request')
    def test_relation_no_unit_data_file(self, request):
        request.method = 'GET'

        mod = relations.Module()

        os.mkdir(os.path.join(self.tempdir, 'myrelation'))
        os.mkdir(os.path.join(self.tempdir, 'myrelation', '1'))

        with self.assertRaises(cherrypy.HTTPError) as e:
            mod.relation(relation='myrelation', unit='1')
            self.assertEqual(e.status, 404)

    @patch('modules.relations.cherrypy.request')
    def test_relation_with_unit(self, request):
        request.method = 'GET'

        mod = relations.Module()

        os.mkdir(os.path.join(self.tempdir, 'myrelation'))
        os.mkdir(os.path.join(self.tempdir, 'myrelation', '1'))
        with open(os.path.join(
                self.tempdir, 'myrelation', '1', 'data'), 'w') as f:
            json.dump({'unit_data_key': 'unit_data_value'}, f)

        expected = {'unit_data_key': 'unit_data_value'}
        self.assertEqual(
            expected,
            json.loads(mod.relation(relation='myrelation', unit='1').decode()))

    def test_list_units(self):
        mod = relations.Module()

        os.mkdir(os.path.join(self.tempdir, 'myrelation'))
        os.mkdir(os.path.join(self.tempdir, 'myrelation', '1'))

        self.assertEqual(['1'], mod.list_units('myrelation'))


class TestRun(TestCase):
    @patch('modules.run.cherrypy.request')
    def test_run_invalid_http_method(self, request):
        request.method = 'GET'

        mod = run.Module()

        with self.assertRaises(cherrypy.HTTPError) as e:
            mod.run()
            self.assertEqual(e.status, 405)

    @patch('modules.run.cherrypy.request')
    def test_run_success(self, request):
        request.method = 'POST'
        request.body.read.return_value = b'echo hello'

        mod = run.Module()

        expected = {'code': 0, 'output': 'hello'}
        self.assertEqual(expected, json.loads(mod.run().decode()))

    @patch('modules.run.cherrypy.response')
    @patch('modules.run.cherrypy.request')
    def test_run_error(self, request, response):
        request.method = 'POST'
        request.body.read.return_value = b'false'

        mod = run.Module()

        expected = {'code': 1, 'output': ''}
        self.assertEqual(expected, json.loads(mod.run().decode()))
        self.assertEqual(500, response.status)


class TestJsonRpc(TestCase):
    def setUp(self):
        class Fixture(object):
            def _validate(self, auth):
                return auth == 'true'

            def test_func(self, *args, **kwargs):
                return dict(args=list(args), kwargs=kwargs)
        self.fixture = Fixture
        self.expected = {'args': [1], 'kwargs': {'b': 2}}

    @patch('json_rpc.cherrypy.request')
    def test_secure_auth_succeeds(self, request):
        request.headers = {'Authorization': 'true'}

        self.fixture.test_func = json_rpc.secure(self.fixture.test_func)
        self.assertEqual(self.expected, self.fixture().test_func(1, b=2))

    @patch('json_rpc.cherrypy.response')
    @patch('json_rpc.cherrypy.request')
    def test_secure_auth_fails(self, request, response):
        request.headers = {'Authorization': 'false'}

        self.fixture.test_func = json_rpc.secure(self.fixture.test_func)
        with self.assertRaises(cherrypy.HTTPError) as e:
            self.fixture().test_func(1, b=2)
            self.assertEqual(e.status, 401)
            self.assertEqual(e.msg, 'Authorization Required')
            self.assertEqual(
                response.headers['WWW-Authenticated'],
                'Basic realm="Secure Area"')

    @patch('json_rpc.cherrypy.response')
    def test_jsonify(self, response):
        response.headers = {}
        self.fixture.test_func = json_rpc.jsonify(self.fixture.test_func)

        self.assertEqual(
            self.expected,
            json.loads(self.fixture().test_func(1, b=2).decode()))
        self.assertEqual(response.headers['Content-Type'], 'application/json')

    @patch('json_rpc.cherrypy.response')
    def test_dont_jsonify(self, response):
        response.headers = {}
        self.fixture.test_func = lambda self: 'test'
        self.fixture.test_func = json_rpc.dont_jsonify(self.fixture.test_func)

        self.assertEqual(b'test', self.fixture().test_func())
        self.assertEqual(response.headers['Content-Type'], 'application/json')



"""Unit test for amulet.wait"""

import os
import unittest
import yaml
from amulet import wait
from amulet import waiter

from amulet.helpers import TimeoutError, JujuVersion
from .helper import JujuStatus

from mock import patch


class WaiterTest(unittest.TestCase):
    @patch('amulet.waiter.juju')
    def test_get_pyjuju_status(self, mock_check_output):
        mstatus = JujuStatus('juju')
        mstatus.add('wordpress')
        mstatus.add('mysql', state='pending')
        mock_check_output.return_value = str(mstatus)

        status = waiter._get_pyjuju_status('dummy')
        self.assertEqual(yaml.safe_load(str(mstatus)), status)
        mock_check_output.assert_called_with(['status', '-e', 'dummy'])

    @patch('amulet.waiter.juju')
    def test_get_pyjuju_status_timeout(self, mj):
        mj.side_effect = [TimeoutError]
        self.assertRaises(TimeoutError, waiter._get_pyjuju_status)

    @patch('amulet.waiter.juju')
    def test_get_pyjuju_status_error(self, mj):
        mj.side_effect = [OSError(0, 'Non-zero exit')]
        self.assertRaises(Exception, waiter._get_pyjuju_status)

    @patch.object(waiter, '_get_pyjuju_status')
    def test_get_gojuju_status(self, mock_pyjuju_status):
        waiter._get_gojuju_status('dummy')
        mock_pyjuju_status.assert_called_with('dummy')

    def test_parse_unit_state(self):
        data = [{'life': 'dying'},
                {'agent-state': 'started',
                 'relations-error': {'key': ['unit']}},
                {'agent-state': 'pending'}]

        self.assertEqual('dying', waiter.get_state(data[0]))
        self.assertEqual(str(data[1]['relations-error']),
                         waiter.get_state(data[1]))
        self.assertEqual('pending', waiter.get_state(data[2]))

    @patch.object(waiter, 'status')
    @patch('amulet.helpers.JujuVersion')
    def test_state_py(self, jver, pyjuju_status):
        jver.side_effect = [JujuVersion(0, 7, 0, False)]
        mstatus = JujuStatus('juju')
        mstatus.add('test-charm')

        pyjuju_status.return_value = yaml.safe_load(str(mstatus))
        output = {'test-charm': {'0': 'started'}}
        self.assertEqual(output, waiter.state(juju_env='test'))

    @patch.object(waiter, 'status')
    @patch('amulet.helpers.JujuVersion')
    def test_state_go(self, jver, pyjuju_status):
        jver.side_effect = [JujuVersion(1, 11, 0, False)]
        mstatus = JujuStatus('juju')
        mstatus.add('test-charm')

        pyjuju_status.return_value = yaml.safe_load(str(mstatus))
        output = {'test-charm': {'0': 'started'}}
        self.assertEqual(output, waiter.state(juju_env='test'))

    @patch.object(waiter, 'status')
    @patch('amulet.helpers.JujuVersion')
    def test_state_specific_units(self, jver, pyjuju_status):
        jver.side_effect = [JujuVersion(0, 7, 0, False)]
        mstatus = JujuStatus('juju')
        mstatus.add('test-charm')
        mstatus.add('test-charm', state='pending')
        mstatus.add('test-charm')
        mstatus.add('test-charm-b')
        mstatus.add('test-charm-b')

        pyjuju_status.return_value = yaml.safe_load(str(mstatus))
        output = {'test-charm': {'1': 'pending'},
                  'test-charm-b': {'0': 'started', '1': 'started'}}
        self.assertEqual(output, waiter.state('test-charm/1', 'test-charm-b',
                                              juju_env='test'))

    @patch.object(waiter, 'status')
    @patch('amulet.helpers.JujuVersion')
    def test_state_subordinate_removal(self, jver, ms):
        jver.side_effect = [JujuVersion(1, 17, 0, False)]
        mstatus = JujuStatus('juju')
        mstatus.add('test-sub')
        mstatus.add('test-srv')
        mstatus.status['services']['test-sub'].pop('units', None)
        mstatus.status['services']['test-sub']['subordinate-to'] = 'test-srv'


        ms.return_value = yaml.safe_load(str(mstatus))
        output = {'test-srv': {'0': 'started'}}
        self.assertEqual(output, waiter.state(juju_env='test'))

    @patch.object(waiter, 'status')
    @patch('amulet.helpers.JujuVersion')
    def test_state_service_not_there(self, jver, pyjuju_status):
        jver.side_effect = [JujuVersion(0, 7, 0, False)]
        mstatus = JujuStatus('juju')

        pyjuju_status.return_value = yaml.safe_load(str(mstatus))
        self.assertRaises(ValueError, waiter.state, 'srvc', juju_env='test')

    @patch.object(waiter, 'status')
    @patch('amulet.helpers.JujuVersion')
    def test_state_timeout(self, jver, pyjuju_status):
        jver.side_effect = [JujuVersion(0, 7, 0, False)]

        pyjuju_status.side_effect = TimeoutError
        self.assertRaises(TimeoutError, waiter.state, juju_env='test')

    @patch.object(waiter, 'status')
    @patch('amulet.helpers.JujuVersion')
    def test_status_error(self, jver, pyjuju_status):
        jver.side_effect = [JujuVersion(0, 7, 0, False)]

        pyjuju_status.side_effect = Exception
        self.assertEqual({}, waiter.state(juju_env='test'))

    @patch('amulet.helpers.JujuVersion')
    def test_status_no_juju_env(self, jver):
        jver.side_effect = [JujuVersion(0, 7, 0, False)]

        self.assertRaises(KeyError, waiter.state)


class WaitTest(unittest.TestCase):
    @patch('amulet.waiter.state')
    def test_wait(self, waiter_status):
        waiter_status.return_value = {'test': {'0': 'started'}}

        self.assertTrue(wait(juju_env='dummy', timeout=1))
        waiter_status.assert_called_with(juju_env='dummy')

    @patch('amulet.waiter.raise_for_state')
    @patch('amulet.timeout')
    def test_wait_timeout(self, tout, waiter_status):
        waiter_status.side_effect = TimeoutError
        tout.side_effect = TimeoutError
        self.assertRaises(TimeoutError, wait, juju_env='dummy')

    @patch('amulet.waiter.state')
    @patch.dict('os.environ', {'JUJU_ENV': 'testing-env'})
    def test_wait_juju_env(self, waiter_status):
        waiter_status.return_value = {'test': {'0': 'started'}}
        wait()
        waiter_status.assert_called_with(juju_env='testing-env')

    @patch('amulet.waiter.state')
    @patch('amulet.default_environment')
    @patch.dict('os.environ', {'JUJU_HOME': '/tmp/juju-home'})
    def test_wait_default_juju_env(self, default_env, waiter_status):
        if not os.path.exists('/tmp/juju-home'):
            os.makedirs('/tmp/juju-home')
        envyaml = yaml.dump({'default': 'testing-default-env',
                             'environments': {'testing-default-env':
                                              {'type': 'local'}}})
        with open('/tmp/juju-home/environments.yaml', 'w') as f:
            f.write(envyaml)

        waiter_status.return_value = {'test': {'0': 'started'}}
        default_env.return_value = 'testing-default-env'
        try:
            wait()
        finally:
            os.remove('/tmp/juju-home/environments.yaml')
            os.removedirs('/tmp/juju-home')

        waiter_status.assert_called_with(juju_env='testing-default-env')

    @patch('amulet.waiter.state')
    def test_wait_not_ready(self, waiter_status):
        waiter_status.side_effect = [{'test': {'0': 'pending'}},
                                     {'test': {'0': 'started'}}]

        self.assertTrue(wait(juju_env='dummy', timeout=1))

    @patch('amulet.waiter.state')
    def test_wait_exception(self, waiter_status):
        waiter_status.side_effect = [Exception,
                                     TimeoutError]

        self.assertRaises(TimeoutError, wait, juju_env='dummy')


class StatusTest(unittest.TestCase):
    @patch('amulet.waiter._get_gojuju_status')
    def test_status_go(self, mpy):
        waiter.status('gojuju')
        mpy.assert_called_with('gojuju')

    @patch('amulet.waiter._get_pyjuju_status')
    @patch('amulet.waiter.JujuVersion')
    def test_status_py(self, mj, mpy):
        mj.side_effect = [JujuVersion(0, 7, 0, False)]
        waiter.status('pyjuju')
        mpy.assert_called_with('pyjuju')

    def test_status_noenv(self):
        self.assertRaises(Exception, waiter.status)

    @patch('amulet.waiter._get_gojuju_status')
    def test_wait_timeout(self, mpy):
        mpy.side_effect = [TimeoutError]
        self.assertRaises(TimeoutError, waiter.status, 'godummy')

    @patch('amulet.waiter._get_gojuju_status')
    def test_wait_exception(self, mpy):
        mpy.side_effect = [Exception]
        self.assertRaises(Exception, waiter.status, 'godummy')




import os
import json
import glob
import json_rpc
import cherrypy

RELATIONS_DIR = '/opt/relations'


class Module (object):
    @json_rpc.expose_anonymous
    def relations(self):
        if not cherrypy.request.method == "GET":
            raise cherrypy.HTTPError(405)

        if not os.path.exists(RELATIONS_DIR):
            return {}

        relations = {}
        for relation in glob.glob(RELATIONS_DIR + '/*'):
            relname = os.path.basename(relation)
            relations[relname] = self.list_units(relname)

        return relations

    @json_rpc.expose_anonymous
    def relation(self, relation=None, unit=None):
        if not cherrypy.request.method == 'GET':
            raise cherrypy.HTTPError(405)

        if not relation:
            raise cherrypy.HTTPError(400)

        if not os.path.exists(os.path.join(RELATIONS_DIR, relation)):
            raise cherrypy.HTTPError(404)

        if not unit:
            data = {}
            for unit in self.list_units(relation):
                data[unit] = json.loads(self.relation(relation, unit).decode())
            return data

        data_file = os.path.join(RELATIONS_DIR, relation, unit, 'data')
        if not os.path.exists(data_file):
            raise cherrypy.HTTPError(404)

        with open(data_file, 'r') as u:
            unit_data = json.loads(u.read())

        return unit_data

    def list_units(self, relation):
        units = []
        for relunits in glob.glob(RELATIONS_DIR + '/%s/*' % relation):
            units.append(os.path.basename(relunits))

        return units




import cherrypy
import functools
import json


def secure(function):
    @functools.wraps(function)
    def wrapped(self, *args, **kwargs):
        auth = cherrypy.request.headers.get("Authorization")
        if auth is None or not self._validate(auth):
            cherrypy.response.headers["WWW-Authenticate"] = (
                "Basic realm=\"Secure Area\"")
            raise cherrypy.HTTPError(401, "Authorization Required")
        else:
            return function(self, *args, **kwargs)
    return wrapped


def jsonify(function):
    @functools.wraps(function)
    def wrapped(*args, **kwargs):
        cherrypy.response.headers['Content-Type'] = 'application/json'
        return json.dumps(function(*args, **kwargs), indent=4).encode()
    return wrapped


def dont_jsonify(function):
    @functools.wraps(function)
    def wrapped(*args, **kwargs):
        cherrypy.response.headers['Content-Type'] = 'application/json'
        return function(*args, **kwargs).encode()
    return wrapped


def expose_secure(function):
    return secure(cherrypy.expose(jsonify(function)))


def expose_secure_dont_jsonify(function):
    return secure(cherrypy.expose(dont_jsonify(function)))


def expose_anonymous(function):
    return cherrypy.expose(jsonify(function))


def expose_anonymous_dont_jsonify(function):
    return cherrypy.expose(dont_jsonify(function))
